@@ -184,7 +184,7 @@ def patch_cffi():
 patch_cffi()
 del patch_cffi
 """)
-    
+
 
 def hook_multiprocessing(finder, module):
     module.__globalnames__.add("AuthenticationError")
@@ -249,6 +249,11 @@ def hook_urllib_request(finder, module):
     """
     finder.excludes.append("_scproxy")
 
+def hook_urllib3(finder, module):
+    """urllib3 embeds a copy of six that requires queue.
+    """
+    finder.import_hook("queue")
+
 def hook_pythoncom(finder, module):
     """pythoncom is a Python extension module with .dll extension,
     usually in the windows system directory as pythoncom3X.dll.
@@ -286,15 +291,16 @@ def hook_tkinter(finder, module):
     tk_dir = os.path.join(os.path.dirname(tcl_dir), 'tk{}'.format(TK_VERSION))
     assert os.path.isdir(tk_dir)
     finder.add_datadirectory("lib/tk", tk_dir, recursive=True)
-    finder.set_min_bundle("tkinter", 2)
     if sys.version_info >= (3,6,0):
         finder.import_hook("imp")
+
     # add environment variables that point to the copied paths at runtime
     finder.add_bootcode("""
 def tk_env_paths():
     import os
+    import sys
     import _tkinter
-    basepath = os.path.dirname(_tkinter.__file__)
+    basepath = os.path.dirname(sys.executable)
     tcl_dir = os.path.join(basepath, 'lib', 'tcl')
     tk_dir = os.path.join(basepath, 'lib', 'tk')
     os.environ["TCL_LIBRARY"] = tcl_dir
@@ -343,8 +349,6 @@ def hook_six(finder, module):
     m = SixImporter(finder,
                     None, "six.moves", finder._optimize)
     finder._add_module("six.moves", m)
-    
-
 
 def hook_matplotlib(finder, module):
     """matplotlib requires data files in a 'mpl-data' subdirectory in
@@ -540,7 +544,7 @@ def hook_numpy_core_numerictypes(finder, module):
 def hook_numpy_core(finder, module):
     finder.ignore("numpy.core._dotblas")
     numpy_core_path = os.path.dirname(module.__loader__.path)
-    #add mkl dlls from numpy.core, if present 
+    #add mkl dlls from numpy.core, if present
     from os import listdir
     dlls = [os.path.join(numpy_core_path,mkl)
             for mkl in listdir(numpy_core_path)
@@ -548,7 +552,53 @@ def hook_numpy_core(finder, module):
     for dll in dlls:
         finder.add_dll(dll)
 
-    
+def hook_pandas(finder, module):
+    #pd_lib_path = os.path.join(os.path.dirname(module.__loader__.path), "_libs")
+    #finder.add_datadirectory("mpl-data", mpl_data_path, recursive=True)
+    depth = getattr(finder,"recursion_depth_pandas", 0)
+    if depth==0:
+        finder.recursion_depth_pandas = depth + 1
+        finder.import_hook("pandas._libs.tslibs.base")
+        finder.recursion_depth_pandas = depth
+
+def hook_Crypto(finder, module):
+    """pycryptodome includes compiled libraries as if they were Python C extensions (as .pyd files).
+    However, they are not, as they cannot be imported by Python. Hence, those files should be treated
+    as .dll files. Furthermore, pycryptodome needs to be patched to import those libraries from an external
+    path, as their import mechanism will not work from the zip file nor from the executable."""
+    # copy all the "pyd" files from pycryptodome to the bundle directory with the correct folder structure
+    crypto_path = os.path.dirname(module.__loader__.path)
+    from pathlib import Path
+    for path in Path(crypto_path).rglob('*.pyd'):
+        finder.add_libfile(str(path.relative_to(os.path.dirname(crypto_path))), path)
+
+    # patch pycryptodome to look for its "pyd" files in the bundle directory
+    finder.add_bootcode("""
+def patch_Crypto():
+    import os
+    import sys
+
+    try:
+        import Crypto.Util._file_system
+    except ImportError:
+        # the hook should not work for pycrypto
+        return
+
+    def override_filename(dir_comps, filename):
+        if dir_comps[0] != "Crypto":
+            raise ValueError("Only available for modules under 'Crypto'")
+
+        dir_comps = list(dir_comps[1:]) + [filename]
+        root_lib = os.path.join(os.path.dirname(sys.executable), 'Crypto')
+
+        return os.path.join(root_lib, *dir_comps)
+
+    Crypto.Util._file_system.pycryptodome_filename = override_filename
+
+patch_Crypto()
+del patch_Crypto
+""")
+
 def hook_scipy_special(finder, module):
     #import pdb;pdb.set_trace()
     depth = getattr(finder,"recursion_depth_special",0)
@@ -567,8 +617,8 @@ def hook_scipy_linalg(finder, module):
         finder.import_hook("scipy.linalg.cython_lapack")
         finder.import_hook("scipy.integrate")
         finder.recursion_depth_linalg = depth
-        
-        
+
+
 def hook_scipy_sparse_csgraph(finder, module):
     depth = getattr(finder,"recursion_depth_sparse",0)
     if depth==0:
@@ -584,16 +634,27 @@ def hook_scipy_optimize(finder, module):
         finder.import_hook("scipy.optimize.minpack2")
         finder.recursion_depth_optimize = depth
 
+def hook_shapely(finder, module):
+    import glob
+    import shapely
+    for dll_path in glob.glob(os.path.join(os.path.abspath(os.path.join(os.path.dirname(shapely.__file__), 'DLLs')), '*.dll')):
+        finder.add_dll(dll_path)
+
 def hook__ssl(finder, module):
     """
-    On python 3.7 and above, _ssl.pyd requires additional dll's to load.
+    On Python 3.7 and above, _ssl.pyd requires additional dll's to load.
     Based on code by Sebastian Krause: https://github.com/anthony-tuininga/cx_Freeze/pull/470
     Apparently, even with the new DLL finder system, this hook is still needed on cp37-win32
+
+    On Python 3.7 an above, _ssl is incompatible with bundle_files=0.
     """
     if sys.version_info < (3, 7, 0):
         return
+
     import glob
     for dll_search in ["libcrypto-*.dll", "libssl-*.dll"]:
         for dll_path in glob.glob(os.path.join(sys.base_prefix, "DLLs", dll_search)):
             dll_name = os.path.basename(dll_path)
             finder.add_dll(dll_path)
+
+    finder.set_min_bundle("_ssl", 1)
